# shellcheck shell=bash

function _exam_env_msg_setup() {
    local _log _dlog

    if (( UID == 0 ))
    then
        # system wide logs
        _log_f="${EXAM_ENV_SYS_LOG_FILE}"
        _d_log_f="${EXAM_ENV_SYS_DEBUG_LOG_FILE}"
    else
        _log_f="${EXAM_ENV_USER_LOG_FILE}"
        _d_log_f="${EXAM_ENV_USER_DEBUG_LOG_FILE}"
    fi

    bb_msg_clear_targets
    bb_msg_set_target log_stderr stderr info progpid notrace
    bb_msg_set_target log file:"$_log_f" info:err full notrace
    bb_msg_set_target log_crit file:"$_log_f" crit full trace
    bb_msg_set_target d_log_debug file:"$_d_log_f" debug3:debug1 full notrace
    bb_msg_set_target d_log file:"$_d_log_f" info full trace
}

function _exam_env_validate_mandatory_config() {
    # check for exam env host and path
    [ -n "${EXAM_ENV_SERVER_HOST-}" ] || \
        bb_fatal "validate_config: EXAM_ENV_SERVER_HOST is not net"
    [ -n "${EXAM_ENV_SERVER_DATA_PATH-}" ] || \
        bb_fatal "validate_config: EXAM_ENV_SERVER_DATA_PATH is not net"
    export EXAM_ENV_SERVER_HOST EXAM_ENV_SERVER_DATA_PATH

    # check for local exam-env config dir
    if [ -n "${EXAM_ENV_CLIENT_CONFIG_PATH-}" ]
    then
        [ -d "$EXAM_ENV_CLIENT_CONFIG_PATH" ] || {
            mkdir -p "$EXAM_ENV_CLIENT_CONFIG_PATH" ||
                bb_fatal "validate_config: failed to create" \
                         "EXAM_ENV_CLIENT_CONFIG_PATH create failed:" \
                         "$EXAM_ENV_CLIENT_CONFIG_PATH"
        }
    else
        bb_fatal "validate_config: EXAM_ENV_CLIENT_CONFIG_PATH is not net"
    fi
    export EXAM_ENV_CLIENT_CONFIG_PATH

    # local priv path
    [ -n "${EXAM_ENV_CLIENT_PRIV_PATH-}" ] || \
        bb_fatal "validate_config EXAM_ENV_CLIENT_PRIV_PATH is not set"

    # check for our SSH ID to call exam server
    if [ -n "${EXAM_ENV_CLIENT_SSHID_PATH-}" ]
    then
        [ -r "$EXAM_ENV_CLIENT_SSHID_PATH" ] || \
            bb_fatal "validate_config: sshid not found at:" \
                     "$EXAM_ENV_CLIENT_SSHID_PATH"
    else
        bb_fatal "validate_config: EXAM_ENV_CLIENT_SSHID_PATH is not set"
    fi
    export EXAM_ENV_CLIENT_SSHID_PATH

    # client user/group
    [ -n "${EXAM_ENV_CLIENT_USER-}" ] ||
        bb_fatal "validate_config: EXAM_ENV_CLIENT_USER is not set"
    [ -n "${EXAM_ENV_CLIENT_GROUP-}" ] ||
        bb_fatal "validate_config: EXAM_ENV_CLIENT_GROUP is not set"
    export EXAM_ENV_CLIENT_USER EXAM_ENV_CLIENT_GROUP
}

# set optional config defaults that can be overwritten later
function _exam_env_set_optional_config_defaults() {
    # log file config defaults
    EXAM_ENV_SYS_LOG_FILE=/var/log/exam-env.log
    EXAM_ENV_SYS_DEBUG_LOG_FILE=/var/log/exam-env-debug.log
    export EXAM_ENV_SYS_LOG_FILE EXAM_ENV_SYS_DEBUG_LOG_FILE
    EXAM_ENV_USER_LOG_FILE=~/.exam-env.log
    EXAM_ENV_USER_DEBUG_LOG_FILE=~/.exam-env-debug.log
    export EXAM_ENV_USER_LOG_FILE EXAM_ENV_USER_DEBUG_LOG_FILE
}

function _exam_env_read_config() {
    local _path="$1"

    if [ -r "$_path" ]
    then
        # shellcheck disable=SC1090
        source "$_path" || \
            bb_fatal "read_config: failed to source config from: $_path"
    else
        bb_fatal "read_config: could not find/read config at: $_path"
    fi
}

# exam_env_resolv_name
# this function resolvs this exams name and sets _EXAM_ENV_NAME
function _exam_env_resolv_name() {
    declare -g _EXAM_ENV_NAME
    _EXAM_ENV_NAME=$(awk -F "/" '/^SUSROOT=/ {print $NF}' /etc/sus/sus.conf) \
        || bb_fatal "exam_env_resolv_name: awk sus.conf failed"

    [ -n "$_EXAM_ENV_NAME" ] || \
        bb_fatal "exam_env_resolv_name: could not resolv name from sus.conf"

    export _EXAM_ENV_NAME
}

function _exam_env_examsrv_reachable() {
    exam_env_call_examsrv true || \
        bb_fatal "_exam_env_check_examsrv: failed to reach examsrv:" \
                 "$EXAM_ENV_SERVER_USER@EXAM_ENV_SERVER_HOST"
}

# exam_env_rsync <arg1> [ <arg2> [..]]
# wrapper for rsync to log output
# rsync will be called with all args passed
# on rsync failure it will retry for 5 minutes
# returns the result of rsync
function exam_env_rsync() {
    (( UID == 0 )) || bb_fatal "exam_env_call_examsrv called as non root"

    local _tmp_dir
    _tmp_dir=$(mktemp -d exam-env-call_examsrv.XXXXXXXXX) || \
        bb_fatal "exam_env_call_examsrv: mktemp -d failed"
    local -i _try=0 _retval

    local _retry_str
    if (( EXAM_ENV_SSH_RETRIES == 0 ))
    then
        printf -v _retry_str "will never give up trying"
    else
        printf -v _retry_str "try %i of %i" "$_try" "$EXAM_ENV_SSH_RETRIES"
    fi

    while true
    do
        bb_msg debug "exam_env_rsync: calling: rsync --archive" \
               " --delete-after -e \"ssh -l $EXAM_ENV_SERVER_USER" \
               "${_EXAM_ENV_SSH_ARGS[*]}\" $*"
        if rsync --archive --delete-after \
                 -e "ssh -l $EXAM_ENV_SERVER_USER ${_EXAM_ENV_SSH_ARGS[*]}" \
                 "$@" >&2  2> "$_tmp_dir/rsync.out"
        then
            # escape loop on success
            _retval=0
            break
        else
            _retval=$?
            bb_msg warning "exam_env_rsync: failed to rsync '$*':"
            bb_msg_from_stdin warning "  rsync-log: " < "$_tmp_dir/rsync.out"
        fi

        # check if we want to retry
        _try=$((_try + 1))
        if (( EXAM_ENV_SSH_RETRIES > -1 &&
                  EXAM_ENV_SSH_RETRIES < _try ))
        then
            rm -rf "$_tmp_dir"
            bb_fatal "exam_env_rsync: max retries reached giving up"
        fi

        bb_msg info "exam_env_rsync: sleeping $EXAM_ENV_SSH_RETRY_WAIT_TIME" \
               " seconds before trying again ($_retry_str)"

        # next try
        sleep "$EXAM_ENV_SSH_RETRY_WAIT_TIME"
    done

    rm -rf "$_tmp_dir"

    return "$_retval"
}

# exam_env_call_examsrv <cmd> [ <arg> [..]]
# runs command on exam server
function exam_env_call_examsrv() {
    (( UID == 0 )) || bb_fatal "exam_env_call_examsrv called as non root"

    local _tmp_dir
    _tmp_dir=$(mktemp -d exam-env-call_examsrv.XXXXXXXXX) || \
        bb_fatal "exam_env_call_examsrv: mktemp -d failed"
    local -i _try=0 _retval

    local _retry_str
    if (( EXAM_ENV_SSH_RETRIES == 0 ))
    then
        printf -v _retry_str "will never give up trying"
    else
        printf -v _retry_str "try %i of %i" "$_try" "$EXAM_ENV_SSH_RETRIES"
    fi

    # shellcheck disable=SC2029
    while true
    do
        _retval=0
        bb_msg debug "exam_env_call_examsrv: calling: " \
               ssh -n "${_EXAM_ENV_SSH_ARGS[*]}" \
               -E "$_tmp_dir/ssherrs" \
               -l "$EXAM_ENV_SERVER_USER" \
               "$EXAM_ENV_SERVER_HOST" "$*"
        if ssh -n "${_EXAM_ENV_SSH_ARGS[@]}" \
            -E "$_tmp_dir/ssherrs" \
            -l "$EXAM_ENV_SERVER_USER" \
            "$EXAM_ENV_SERVER_HOST" "$@"
        then
            # escape loop on success
            _retval=0
        else
            _retval="$?"

            # check if its error on other side or ssh problem
            if (( _retval == 255 ))
            then
                if [ -s "$_tmp_dir/ssherrs" ]
                then
                    bb_msg warning "exam_env_call_examsrv: SSH failed:"
                    bb_msg_from_stdin warning "  ssh-log: " \
                                      < "$_tmp_dir/ssherrs"
                    bb_msg warning "exam_env_call_examsrv: sleeping " \
                           "$EXAM_ENV_SSH_RETRY_WAIT_TIME" \
                           " seconds before trying again ($_retry_str)"

                    # check if we want to retry
                    _try=$((_try + 1))
                    if (( EXAM_ENV_SSH_RETRIES > -1 &&
                              EXAM_ENV_SSH_RETRIES < _try ))
                    then
                        rm -rf "$_tmp_dir"
                        bb_fatal "exam_env_call_examsrv: max retries reached" \
                                 "giving up"
                    fi

                    # cleanup for next try
                    rm -f "$_tmp_dir/ssherrs"

                    # next try
                    sleep "$EXAM_ENV_SSH_RETRY_WAIT_TIME"
                    continue
                fi
            fi
        fi

        # if we reached this we are happily out
        break
    done

    rm -rf "$_tmp_dir"

    return "$_retval"
}

# source_script <filename>
# returns the result of the script (its last command)
function exam_env_source_script() {
    (( $# == 1 )) || \
        bb_fatal "exam_env_source_script called with $# args instead of 1"

    local _script="$1"

    if [ -f "$_script" ]
    then
        bb_msg debug "sourcing script $_script .."
        # shellcheck disable=SC1090
        source "$_script" || \
            bb_fatal "exam_env_source_script: failed to source: $_script"
    else
        bb_msg warning "exam_env_source_script:" \
               "no script $_script found, skipping it"
    fi
}

# sync_config pre|full
# synchronizes the config from exam server
# pre is for prerollout and will download only relevant parts of the config
function exam_env_sync_config() {
    (( $# == 1 )) || \
        bb_fatal "exam_env_sync_config called with $# instead of 1 arg"

    local -a _extra_args
    case "$1" in
        pre) _extra_args=(--include 'users/***' --include 'sshkeys/***'
                          --include 'scripts' --include 'scripts/config.sh'
                          --include 'img/***' --exclude='**');;
        full) _extra_args=();;
        *) bb_fatal "exam_env_sync_config with illegal arg: $1"
    esac

    bb_msg info "downloading exam config now.."
    mkdir -p "$EXAM_ENV_CLIENT_CONFIG_PATH"
    exam_env_rsync \
        "${_extra_args[@]}" \
        "$EXAM_ENV_SERVER_HOST:$_EXAM_ENV_SERVER_EXAM_PATH/config/." \
        "$EXAM_ENV_CLIENT_CONFIG_PATH/." || {
        bb_fatal "config fetch rsync failed"
    }
    bb_msg info "successfully downloaded config"

}

# decomment
# a pipe that removes lines that are comments
# shellcheck disable=SC2120
function exam_env_decomment() {
    (( $# == 0 )) || \
        bb_fatal "exam_env_decomment called with $# instead of 0 args"

    grep -E -v '^[[:space:]]*(#.*)?$' || true
}

# check_delay
# checks on exam server if delay is requested, returns 1 on delay, 0 if none
function exam_env_check_delay() {
    # shellcheck disable=SC2119
    if exam_env_call_examsrv "cd \"$_EXAM_ENV_SERVER_EXAM_PATH\"; [ -f DELAY-ROLLOUT ] && cat DELAY-ROLLOUT" | \
            exam_env_decomment | \
         awk '/^[[:space:]]*DELAY[[:space:]]*$/ { exit 1 }'
    then
        bb_msg debug "exam_env_check_delay: no lock active"
        # no lock active
        return 0
    else
        bb_msg debug "exam_env_check_delay:" \
               "global lock active, checking exceptions.."
    fi

    # shellcheck disable=SC2119
    if exam_env_call_examsrv "cd \"$_EXAM_ENV_SERVER_EXAM_PATH\"; [ -f DELAY-EXCEPTIONS ] && cat DELAY-EXCEPTIONS" | \
        exam_env_decomment | \
        awk '/^[[:space:]]*'"$(hostname -s)"'[[:space:]]*$/ { exit 1 }'
    then
        # not found in exceptions
        bb_msg debug "exam_env_check_delay:" \
               "host ($(hostname -s)) not found in exceptions"
        return 1
    fi

    # found in exceptions
    bb_msg info "exam_env_check_delay:" \
           "host $(hostname -s) is in DELAY-EXCEPTIONS, overriding DELAY"
    return 0
}

# exam_env_get_state
# return state, result can be state set by exam_env_change_state or "virgin"
function exam_env_get_state() {
    if [ -f "$_EXAM_ENV_STATE_FILE" ]
    then
        cat "$_EXAM_ENV_STATE_FILE"
    else
        echo "virgin"
    fi
}

# exam_env_get_pretty_state
# return state, result can be state set by exam_env_change_state or "virgin"
function exam_env_get_pretty_state() {
    if [ -f "${_EXAM_ENV_STATE_FILE}.pretty" ]
    then
        cat "${_EXAM_ENV_STATE_FILE}.pretty"
    else
        echo "virgin"
    fi
}
# exam_env_change_state <state>
# where state can be one of:
#  prerollout-{running,ok,error} delayed {rollout,boottime}-{running,ok,error}
function exam_env_change_state() {
    local pretty_state

    local state="$1"

    case "$state" in
        prerollout-running)
            pretty_state="prerollout currently running..";;
        prerollout-ok)
            pretty_state="prerollout okay, state should change soon";;
        prerollout-error)
            pretty_state="prerollout error, this is fatal";;
        delayed)
            pretty_state="delay active, exam rollout not started";;
        rollout-running)
            pretty_state="exam rollout currently running..";;
        rollout-ok)
            pretty_state="exam rollout okay, no boottime done yet";;
        rollout-error)
            pretty_state="exam rollout error, this is fatal";;
        boottime-running)
            pretty_state="boottime currently running..";;
        boottime-ok)
            pretty_state="boottime okay";;
        boottime-error)
            pretty_state="boottime failed, fix scripts and try again";;
        *)
            pretty_state="unknown error, this is fatal";;
    esac

    echo "$state" > "$_EXAM_ENV_STATE_FILE"
    echo "$pretty_state" > "${_EXAM_ENV_STATE_FILE}.pretty"

    sed -ri 's#^PRETTY_NAME="([^)]+\)).*$#PRETTY_NAME="\1 - exam-env '"($_EXAM_ENV_NAME) $pretty_state"'"#' /etc/os-release

    [ -f /etc/motd.orig ] || mv /etc/motd /etc/motd.orig || true
    echo "exam-env:${_EXAM_ENV_NAME-unknown} $state ($pretty_state)" > /etc/motd
    if [ -f /var/local/exam-env-rollout-timestamp ]
    then
        echo "(rollout ran on $(cat /var/local/exam-env-rollout-timestamp)))" \
            >> /etc/motd
    fi

    bb_msg info "exam_env_change_state: state is now $state ($pretty_state)"
}

function exam_env_squelch_console() {
    bb_msg debug "exam_env_squelch_console called"
    echo 6 > "/proc/sys/kernel/printk" || true # try to squelsh console
}

function exam_env_loop_forever() {
    exam_env_squelch_console
    bb_finalize
    bb_msg crit "exam_env_loop_forever: uncorrectable error, giving up" \
        "(press ENTER to reboot)"
    read -r
    exec reboot || {
        bb_msg emerg \
            "exam_env_loop_forever: got ENTER but bb_exec reboot failed"
        exit 1
    }
}

function exam_env_assert_state() {
    local _our_state
    local -a _args=("$@")

    _our_state="$(exam_env_get_state)" || _our_state=UNKNOWN

    while (( $# > 0 ))
    do
        if [ "$_our_state" == "$1" ]
        then
            # we found a good state
            return 0
        fi

        # try next arg, if any
        shift
    done

    bb_msg emerg "exam_env_assert_state: system is in state $_our_state" \
        "but should be in one of: ${_args[*]}"
    exam_env_loop_forever
}

function exam_env_validate_config_sh() {
    (( $# == 1 )) || \
        bb_fatal "exam_env_validate_config_sh called with $# instead of 1 arg"

    [ -f "$1" ] || bb_fatal "exam_env_validate_config_sh: no such file: $1"

    local -i _retval=0
    if grep -qE "^(EXAM_ENV_(SERVER|SSH|CLIENT)|_EXAM_ENV|LVATEST)" "$1"
    then
        bb_msg err "exam_env_validate_config_sh: found illegal lines:"
        grep -qE "^(EXAM_ENV_(SERVER|SSH|CLIENT)|_EXAM_ENV|LVATEST)" "$1" | \
            bb_msg_from_stdin err "  | "
        return 1
    fi
    return "$_retval"
}

# exam_env_re_check_ssh_key  <usename> <pubkey>
# check if ssh key was updated, created or removed
function exam_env_re_check_ssh_key() {
    (( $# == 2 )) || \
        bb_fatal "exam_env_re_check_ssh_key called with $# instead of 2 args"

    local username="$1"
    local pubkey="$2"
    local dst="/home/$username/.ssh/authorized_keys"

    if [ -f "$pubkey" ]
    then
        if [ -f "$dst" ]
        then
            # update changed key
            diff -q "$pubkey" "$dst" || {
                bb_msg debug "exam_env_re_check_ssh_key:" \
                       "updating pubkey for user $username"
                cat "$pubkey" > "$dst"
            }
        else
            # install key
            bb_msg debug "exam_env_re_check_ssh_key:" \
                   "installing new pubkey for user $username"
            mkdir -p "/home/$username/.ssh"
            cp "$pubkey" "$dst"
            chown -R "$username:$username" "/home/$username/.ssh"
            chmod -R go-rwx "/home/$username/.ssh"
        fi
    else
        # no pubkey in config
        bb_msg warning \
               "exam_env_re_check_ssh_key: user $username has no pubkey :("
        if [ -f "$dst" ]
        then
            bb_msg warning "exam_env_re_check_ssh_key:" \
                   "sadly removing pubkey from user $username"
            rm -f "$dst"
        fi
    fi
}

function exam_env_useradd_no_mail_spool() {
    local -i _retval

    # change settings so no mail spool entry is created
    cp -a /etc/default/useradd /etc/default/useradd.before-exam-env
    sed -i -r 's#^CREATE_MAIL_SPOOL=yes#CREATE_MAIL_SPOOL=no#' \
        /etc/default/useradd

    if useradd "$@"
    then
        _retval=0
    else
        _retval="$?"
    fi

    # retore useradd settings
    mv /etc/default/useradd.before-exam-env /etc/default/useradd

    return "$_retval"
}

function exam_env_create_admin_users() {
    [ -f "$EXAM_ENV_CLIENT_CONFIG_PATH/users/admin-users.csv" ] || {
        bb_msg info "create_admin_users: nothing found, skipping"
        return 0
    }

    local _line USERNAME GECOS PWHASH
    while read -r _line || [ -n "$_line" ]
    do
        # ignore comments and empty lines
        if [[ "$_line" =~ ^[[:space:]]*(#.*)?$ ]]
        then
            bb_msg debug "create_admin_users: skipping comment line: $_line"
            continue
        fi

        local _line_ar
        IFS=\; read -r -a _line_ar <<< "$_line"

        [[ ${#_line_ar[@]} == 3 ]] || bb_fatal "create_admin_users:" \
            "illegal line in admin-users.csv: $_line"

        bb_msg debug "create_admin_users: processing line: ${_line_ar[*]}"

        USERNAME=${_line_ar[0]}
        GECOS=${_line_ar[1]}
        PWHASH=${_line_ar[2]}

        if getent passwd "$USERNAME" &> /dev/null
        then
            bb_msg info "modifying existing user $USERNAME"
            usermod --password "$PWHASH" \
                    --comment "$GECOS" "$USERNAME"
            exam_env_re_check_ssh_key \
                "$USERNAME" \
                "$EXAM_ENV_CLIENT_CONFIG_PATH/sshkeys/$USERNAME.pub"
        else
            bb_msg info "create_admin_users: adding '$USERNAME'" \
                   "gecos='$GECOS ($_EXAM_ENV_NAME admin user)'"
            exam-env-add-admin-user \
                "$USERNAME" \
                "$GECOS ($_EXAM_ENV_NAME admin user)" \
                "$PWHASH" \
                "$EXAM_ENV_CLIENT_CONFIG_PATH/sshkeys/$USERNAME.pub"
        fi
    done < "$EXAM_ENV_CLIENT_CONFIG_PATH/users/admin-users.csv"
}

function _exam_env_internal_var_init() {
    # note: _EXAM_ENV_NAME is set by _exam_env_resolv_name

    # local file to persist state
    declare -g _EXAM_ENV_STATE_FILE="/var/local/exam-env-state"
    export _EXAM_ENV_STATE_FILE

    # shortcut path to our exam at exam server
    declare -g _EXAM_ENV_SERVER_EXAM_PATH
    _EXAM_ENV_SERVER_EXAM_PATH="$EXAM_ENV_SERVER_DATA_PATH/$_EXAM_ENV_NAME"
    export _EXAM_ENV_SERVER_EXAM_PATH

    # args that always will be passed to ssh
    declare -g -a _EXAM_ENV_SSH_ARGS
    _EXAM_ENV_SSH_ARGS=(
        -x
        -a
        -k
        -C
        -T
        -i "$EXAM_ENV_CLIENT_SSHID_PATH"
        -oConnectTimeout="${EXAM_ENV_SSH_CONNECT_TIMEOUT-5}"
        -oPasswordAuthentication=no
    )
}

# shellcheck disable=SC1091
source /usr/libexec/bash-base.bash || echo BLAH

function _exam_env_common_bootstrap() {
    # make sure bash-base is loaded
    if [ -z "${BASH_BASE_IS_LOADED-}" ]
    then
        echo "$0: bash-base must be loaded, aborting" >&2
        exit 1
    fi

    # check for bash-base version
    bb_version_check_loaded 0.6.4 || {
        echo "$0: bash-base version is too old (< 0.6.4)" >&2
        exit 1
    }

    # set optional config defaults
    _exam_env_set_optional_config_defaults

    # bootstrap msg setup to record early errors
    _exam_env_msg_setup

    # read config file
    _exam_env_read_config "/etc/default/exam-env-client"

    # second msg setup to honor config values
    _exam_env_msg_setup

    # make sure all mandatory config options are present
    _exam_env_validate_mandatory_config

    # resolv exams name from SUS config
    _exam_env_resolv_name

    # initialze internal variables
    _exam_env_internal_var_init

    # check if its possible to reach examsrv
    _exam_env_examsrv_reachable
}

_exam_env_common_bootstrap
